# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import os
import random
from paddle.io import Dataset
import json

from .imaug import transform, create_operators


class PubTabDataSet(Dataset):
    def __init__(self, config, mode, logger, seed=None):
        super(PubTabDataSet, self).__init__()
        self.logger = logger

        global_config = config['Global']
        dataset_config = config[mode]['dataset']
        loader_config = config[mode]['loader']

        label_file_path = dataset_config.pop('label_file_path')

        self.data_dir = dataset_config['data_dir']
        self.do_shuffle = loader_config['shuffle']
        self.do_hard_select = False
        if 'hard_select' in loader_config:
            self.do_hard_select = loader_config['hard_select']
            self.hard_prob = loader_config['hard_prob']
        if self.do_hard_select:
            self.img_select_prob = self.load_hard_select_prob()
        self.table_select_type = None
        if 'table_select_type' in loader_config:
            self.table_select_type = loader_config['table_select_type']
            self.table_select_prob = loader_config['table_select_prob']

        self.seed = seed
        logger.info("Initialize indexs of datasets:%s" % label_file_path)
        with open(label_file_path, "rb") as f:
            self.data_lines = f.readlines()
        self.data_idx_order_list = list(range(len(self.data_lines)))
        if mode.lower() == "train":
            self.shuffle_data_random()
        self.ops = create_operators(dataset_config['transforms'], global_config)

    def shuffle_data_random(self):
        if self.do_shuffle:
            random.seed(self.seed)
            random.shuffle(self.data_lines)
        return

    def __getitem__(self, idx):
        try:
            data_line = self.data_lines[idx]
            data_line = data_line.decode('utf-8').strip("\n")
            info = json.loads(data_line)
            file_name = info['filename']
            select_flag = True
            if self.do_hard_select:
                prob = self.img_select_prob[file_name]
                if prob < random.uniform(0, 1):
                    select_flag = False
            
            if self.table_select_type:
                structure = info['html']['structure']['tokens'].copy()
                structure_str = ''.join(structure)
                table_type = "simple"
                if 'colspan' in structure_str or 'rowspan' in structure_str:
                    table_type = "complex"
                if table_type == "complex":
                    if self.table_select_prob < random.uniform(0, 1):
                        select_flag = False                    
            
            if select_flag:
                cells = info['html']['cells'].copy()
                structure = info['html']['structure'].copy()
                img_path = os.path.join(self.data_dir, file_name)
                data = {'img_path': img_path, 'cells': cells, 'structure':structure}
                if not os.path.exists(img_path):
                    raise Exception("{} does not exist!".format(img_path))
                with open(data['img_path'], 'rb') as f:
                    img = f.read()
                    data['image'] = img
                outs = transform(data, self.ops)
            else:
                outs = None
        except Exception as e:
            self.logger.error(
                "When parsing line {}, error happened with msg: {}".format(
                    data_line, e))
            outs = None
        if outs is None:
            return self.__getitem__(np.random.randint(self.__len__()))
        return outs

    def __len__(self):
        return len(self.data_idx_order_list)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import importlib

from .base_model import BaseModel
from .distillation_model import DistillationModel

__all__ = ['build_model']


def build_model(config):
    config = copy.deepcopy(config)
    if not "name" in config:
        arch = BaseModel(config)
    else:
        name = config.pop("name")
        mod = importlib.import_module(__name__)
        arch = getattr(mod, name)(config)
    return arch



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from: 
https://github.com/open-mmlab/mmocr/blob/main/mmocr/models/textrecog/encoders/sar_encoder.py
https://github.com/open-mmlab/mmocr/blob/main/mmocr/models/textrecog/decoders/sar_decoder.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import paddle
from paddle import ParamAttr
import paddle.nn as nn
import paddle.nn.functional as F


class SAREncoder(nn.Layer):
    """
    Args:
        enc_bi_rnn (bool): If True, use bidirectional RNN in encoder.
        enc_drop_rnn (float): Dropout probability of RNN layer in encoder.
        enc_gru (bool): If True, use GRU, else LSTM in encoder.
        d_model (int): Dim of channels from backbone.
        d_enc (int): Dim of encoder RNN layer.
        mask (bool): If True, mask padding in RNN sequence.
    """

    def __init__(self,
                 enc_bi_rnn=False,
                 enc_drop_rnn=0.1,
                 enc_gru=False,
                 d_model=512,
                 d_enc=512,
                 mask=True,
                 **kwargs):
        super().__init__()
        assert isinstance(enc_bi_rnn, bool)
        assert isinstance(enc_drop_rnn, (int, float))
        assert 0 <= enc_drop_rnn < 1.0
        assert isinstance(enc_gru, bool)
        assert isinstance(d_model, int)
        assert isinstance(d_enc, int)
        assert isinstance(mask, bool)

        self.enc_bi_rnn = enc_bi_rnn
        self.enc_drop_rnn = enc_drop_rnn
        self.mask = mask

        # LSTM Encoder
        if enc_bi_rnn:
            direction = 'bidirectional'
        else:
            direction = 'forward'
        kwargs = dict(
            input_size=d_model,
            hidden_size=d_enc,
            num_layers=2,
            time_major=False,
            dropout=enc_drop_rnn,
            direction=direction)
        if enc_gru:
            self.rnn_encoder = nn.GRU(**kwargs)
        else:
            self.rnn_encoder = nn.LSTM(**kwargs)

        # global feature transformation
        encoder_rnn_out_size = d_enc * (int(enc_bi_rnn) + 1)
        self.linear = nn.Linear(encoder_rnn_out_size, encoder_rnn_out_size)

    def forward(self, feat, img_metas=None):
        if img_metas is not None:
            assert len(img_metas[0]) == feat.shape[0]

        valid_ratios = None
        if img_metas is not None and self.mask:
            valid_ratios = img_metas[-1]

        h_feat = feat.shape[2]  # bsz c h w
        feat_v = F.max_pool2d(
            feat, kernel_size=(h_feat, 1), stride=1, padding=0)
        feat_v = feat_v.squeeze(2)  # bsz * C * W
        feat_v = paddle.transpose(feat_v, perm=[0, 2, 1])  # bsz * W * C
        holistic_feat = self.rnn_encoder(feat_v)[0]  # bsz * T * C

        if valid_ratios is not None:
            valid_hf = []
            T = holistic_feat.shape[1]
            for i, valid_ratio in enumerate(valid_ratios):
                valid_step = min(T, math.ceil(T * valid_ratio)) - 1
                valid_hf.append(holistic_feat[i, valid_step, :])
            valid_hf = paddle.stack(valid_hf, axis=0)
        else:
            valid_hf = holistic_feat[:, -1, :]  # bsz * C
        holistic_feat = self.linear(valid_hf)  # bsz * C

        return holistic_feat


class BaseDecoder(nn.Layer):
    def __init__(self, **kwargs):
        super().__init__()

    def forward_train(self, feat, out_enc, targets, img_metas):
        raise NotImplementedError

    def forward_test(self, feat, out_enc, img_metas):
        raise NotImplementedError

    def forward(self,
                feat,
                out_enc,
                label=None,
                img_metas=None,
                train_mode=True):
        self.train_mode = train_mode

        if train_mode:
            return self.forward_train(feat, out_enc, label, img_metas)
        return self.forward_test(feat, out_enc, img_metas)


class ParallelSARDecoder(BaseDecoder):
    """
    Args:
        out_channels (int): Output class number.
        enc_bi_rnn (bool): If True, use bidirectional RNN in encoder.
        dec_bi_rnn (bool): If True, use bidirectional RNN in decoder.
        dec_drop_rnn (float): Dropout of RNN layer in decoder.
        dec_gru (bool): If True, use GRU, else LSTM in decoder.
        d_model (int): Dim of channels from backbone.
        d_enc (int): Dim of encoder RNN layer.
        d_k (int): Dim of channels of attention module.
        pred_dropout (float): Dropout probability of prediction layer.
        max_seq_len (int): Maximum sequence length for decoding.
        mask (bool): If True, mask padding in feature map.
        start_idx (int): Index of start token.
        padding_idx (int): Index of padding token.
        pred_concat (bool): If True, concat glimpse feature from
            attention with holistic feature and hidden state.
    """

    def __init__(
            self,
            out_channels,  # 90 + unknown + start + padding
            enc_bi_rnn=False,
            dec_bi_rnn=False,
            dec_drop_rnn=0.0,
            dec_gru=False,
            d_model=512,
            d_enc=512,
            d_k=64,
            pred_dropout=0.1,
            max_text_length=30,
            mask=True,
            pred_concat=True,
            **kwargs):
        super().__init__()

        self.num_classes = out_channels
        self.enc_bi_rnn = enc_bi_rnn
        self.d_k = d_k
        self.start_idx = out_channels - 2
        self.padding_idx = out_channels - 1
        self.max_seq_len = max_text_length
        self.mask = mask
        self.pred_concat = pred_concat

        encoder_rnn_out_size = d_enc * (int(enc_bi_rnn) + 1)
        decoder_rnn_out_size = encoder_rnn_out_size * (int(dec_bi_rnn) + 1)

        # 2D attention layer
        self.conv1x1_1 = nn.Linear(decoder_rnn_out_size, d_k)
        self.conv3x3_1 = nn.Conv2D(
            d_model, d_k, kernel_size=3, stride=1, padding=1)
        self.conv1x1_2 = nn.Linear(d_k, 1)

        # Decoder RNN layer
        if dec_bi_rnn:
            direction = 'bidirectional'
        else:
            direction = 'forward'

        kwargs = dict(
            input_size=encoder_rnn_out_size,
            hidden_size=encoder_rnn_out_size,
            num_layers=2,
            time_major=False,
            dropout=dec_drop_rnn,
            direction=direction)
        if dec_gru:
            self.rnn_decoder = nn.GRU(**kwargs)
        else:
            self.rnn_decoder = nn.LSTM(**kwargs)

        # Decoder input embedding
        self.embedding = nn.Embedding(
            self.num_classes,
            encoder_rnn_out_size,
            padding_idx=self.padding_idx)

        # Prediction layer
        self.pred_dropout = nn.Dropout(pred_dropout)
        pred_num_classes = self.num_classes - 1
        if pred_concat:
            fc_in_channel = decoder_rnn_out_size + d_model + d_enc
        else:
            fc_in_channel = d_model
        self.prediction = nn.Linear(fc_in_channel, pred_num_classes)

    def _2d_attention(self,
                      decoder_input,
                      feat,
                      holistic_feat,
                      valid_ratios=None):

        y = self.rnn_decoder(decoder_input)[0]
        # y: bsz * (seq_len + 1) * hidden_size

        attn_query = self.conv1x1_1(y)  # bsz * (seq_len + 1) * attn_size
        bsz, seq_len, attn_size = attn_query.shape
        attn_query = paddle.unsqueeze(attn_query, axis=[3, 4])
        # (bsz, seq_len + 1, attn_size, 1, 1)

        attn_key = self.conv3x3_1(feat)
        # bsz * attn_size * h * w
        attn_key = attn_key.unsqueeze(1)
        # bsz * 1 * attn_size * h * w

        attn_weight = paddle.tanh(paddle.add(attn_key, attn_query))

        # bsz * (seq_len + 1) * attn_size * h * w
        attn_weight = paddle.transpose(attn_weight, perm=[0, 1, 3, 4, 2])
        # bsz * (seq_len + 1) * h * w * attn_size
        attn_weight = self.conv1x1_2(attn_weight)
        # bsz * (seq_len + 1) * h * w * 1
        bsz, T, h, w, c = attn_weight.shape
        assert c == 1

        if valid_ratios is not None:
            # cal mask of attention weight
            for i, valid_ratio in enumerate(valid_ratios):
                valid_width = min(w, math.ceil(w * valid_ratio))
                if valid_width < w:
                    attn_weight[i, :, :, valid_width:, :] = float('-inf')

        attn_weight = paddle.reshape(attn_weight, [bsz, T, -1])
        attn_weight = F.softmax(attn_weight, axis=-1)

        attn_weight = paddle.reshape(attn_weight, [bsz, T, h, w, c])
        attn_weight = paddle.transpose(attn_weight, perm=[0, 1, 4, 2, 3])
        # attn_weight: bsz * T * c * h * w
        # feat: bsz * c * h * w
        attn_feat = paddle.sum(paddle.multiply(feat.unsqueeze(1), attn_weight),
                               (3, 4),
                               keepdim=False)
        # bsz * (seq_len + 1) * C

        # Linear transformation
        if self.pred_concat:
            hf_c = holistic_feat.shape[-1]
            holistic_feat = paddle.expand(
                holistic_feat, shape=[bsz, seq_len, hf_c])
            y = self.prediction(paddle.concat((y, attn_feat, holistic_feat), 2))
        else:
            y = self.prediction(attn_feat)
        # bsz * (seq_len + 1) * num_classes
        if self.train_mode:
            y = self.pred_dropout(y)

        return y

    def forward_train(self, feat, out_enc, label, img_metas):
        '''
        img_metas: [label, valid_ratio]
        '''
        if img_metas is not None:
            assert len(img_metas[0]) == feat.shape[0]

        valid_ratios = None
        if img_metas is not None and self.mask:
            valid_ratios = img_metas[-1]

        lab_embedding = self.embedding(label)
        # bsz * seq_len * emb_dim
        out_enc = out_enc.unsqueeze(1)
        # bsz * 1 * emb_dim
        in_dec = paddle.concat((out_enc, lab_embedding), axis=1)
        # bsz * (seq_len + 1) * C
        out_dec = self._2d_attention(
            in_dec, feat, out_enc, valid_ratios=valid_ratios)
        # bsz * (seq_len + 1) * num_classes

        return out_dec[:, 1:, :]  # bsz * seq_len * num_classes

    def forward_test(self, feat, out_enc, img_metas):
        if img_metas is not None:
            assert len(img_metas[0]) == feat.shape[0]

        valid_ratios = None
        if img_metas is not None and self.mask:
            valid_ratios = img_metas[-1]

        seq_len = self.max_seq_len
        bsz = feat.shape[0]
        start_token = paddle.full(
            (bsz, ), fill_value=self.start_idx, dtype='int64')
        # bsz
        start_token = self.embedding(start_token)
        # bsz * emb_dim
        emb_dim = start_token.shape[1]
        start_token = start_token.unsqueeze(1)
        start_token = paddle.expand(start_token, shape=[bsz, seq_len, emb_dim])
        # bsz * seq_len * emb_dim
        out_enc = out_enc.unsqueeze(1)
        # bsz * 1 * emb_dim
        decoder_input = paddle.concat((out_enc, start_token), axis=1)
        # bsz * (seq_len + 1) * emb_dim

        outputs = []
        for i in range(1, seq_len + 1):
            decoder_output = self._2d_attention(
                decoder_input, feat, out_enc, valid_ratios=valid_ratios)
            char_output = decoder_output[:, i, :]  # bsz * num_classes
            char_output = F.softmax(char_output, -1)
            outputs.append(char_output)
            max_idx = paddle.argmax(char_output, axis=1, keepdim=False)
            char_embedding = self.embedding(max_idx)  # bsz * emb_dim
            if i < seq_len:
                decoder_input[:, i + 1, :] = char_embedding

        outputs = paddle.stack(outputs, 1)  # bsz * seq_len * num_classes

        return outputs


class SARHead(nn.Layer):
    def __init__(self,
                 out_channels,
                 enc_bi_rnn=False,
                 enc_drop_rnn=0.1,
                 enc_gru=False,
                 dec_bi_rnn=False,
                 dec_drop_rnn=0.0,
                 dec_gru=False,
                 d_k=512,
                 pred_dropout=0.1,
                 max_text_length=30,
                 pred_concat=True,
                 **kwargs):
        super(SARHead, self).__init__()

        # encoder module
        self.encoder = SAREncoder(
            enc_bi_rnn=enc_bi_rnn, enc_drop_rnn=enc_drop_rnn, enc_gru=enc_gru)

        # decoder module
        self.decoder = ParallelSARDecoder(
            out_channels=out_channels,
            enc_bi_rnn=enc_bi_rnn,
            dec_bi_rnn=dec_bi_rnn,
            dec_drop_rnn=dec_drop_rnn,
            dec_gru=dec_gru,
            d_k=d_k,
            pred_dropout=pred_dropout,
            max_text_length=max_text_length,
            pred_concat=pred_concat)

    def forward(self, feat, targets=None):
        '''
        img_metas: [label, valid_ratio]
        '''
        holistic_feat = self.encoder(feat, targets)  # bsz c

        if self.training:
            label = targets[0]  # label
            label = paddle.to_tensor(label, dtype='int64')
            final_out = self.decoder(
                feat, holistic_feat, label, img_metas=targets)
        if not self.training:
            final_out = self.decoder(
                feat,
                holistic_feat,
                label=None,
                img_metas=targets,
                train_mode=False)
            # (bsz, seq_len, num_classes)

        return final_out



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/4

""" 惟海法师网课考勤工具

为了使用便捷明了，程序开发上
1、使用中文命名
2、牺牲了一定的工程、扩展灵活性，用了尽可能多的自动推导，而不是参数配置
"""
from pyxllib.prog.pupil import check_install_package

check_install_package('fire')  # 自动安装依赖包

from collections import Counter
from datetime import date
import datetime
import math
import os
import re
import time

import fire
import pandas as pd
from tqdm import tqdm

from pyxllib.text.pupil import chinese2digits, grp_chinese_char
from pyxllib.file.xlsxlib import openpyxl
from pyxllib.file.specialist import XlPath
from pyxllib.prog.specialist import parse_datetime, browser, TicToc, dprint
from pyxllib.cv.rgbfmt import RgbFormatter
from pyxllib.data.sqlite import Connection

from pyxllib.ext.seleniumlib import XlChrome


class 网课考勤:
    def __init__(self, today=None):
        self.返款标题 = ''
        self.表格路径 = r'考勤.xlsx'
        self.在线表格 = 'https://docs.qq.com/sheet/DUlF1UnRackJ2Vm5U'  # 生成日报用
        self.开课日期 = '2022-01-08'
        self.视频返款 = [20, 15, 10, 5, 0, 0]  # 直播(当堂)/第1天（当天）/第2天/第3天/第4天/第5天，完成观看的依次返款额。
        self.打卡返款 = {5: 100, 10: 150, 15: 200}  # 打卡满5/10/15次的返款额
        self._init(today)

        self.driver = None  # 浏览器脚本

    def __1_基础数据处理(self):
        pass

    def _init(self, today=None):
        """ 可以手动指定today，适合某些场景下的调试 """
        self.课次数 = len(self.课程链接) - 1  # 课次一般都是固定21课，不用改
        self.达标时长 = 30  # 单位:分钟。在线听课的时长达到此标准以上，才计算为完成一节课的学习
        # 一般是3天，因为回放第1、2、3天都会数额，第4天就没数额了，所以第4天返款
        self.回放返款天数 = sum((x > 0) for x in self.视频返款[1:])

        self.root = XlPath(self.表格路径).parent
        self.wb = openpyxl.load_workbook(self.表格路径)
        self.ws = self.wb['考勤表']
        if today:
            dt = parse_datetime(today)
            self.today = date(dt.year, dt.month, dt.day)
        else:
            self.today = date.today()
        self.开课日期 = date.fromisoformat(self.开课日期)
        # self.觉观禅课 = (self.开课日期.day == 1)
        self.觉观禅课 = '觉观' in self.返款标题
        self.当天课次 = (self.today - self.开课日期).days + 1  # 这是用于逻辑运算的，可能超过实际课次数
        self.当天课次2 = min(self.当天课次, self.课次数)
        self.结束课次 = self.当天课次 - len(self.视频返款) + 1  # 这是用于逻辑运算的，可能有负值
        self.结束课次2 = max(0, self.结束课次)
        self.用户列表 = pd.read_csv(self.get_file('数据表/用户列表导出*.csv'))
        # 用户列表 = 用户列表[用户列表['账号状态'] == '正常']

        self.考勤表出现次数 = Counter()
        for f in self.root.glob('数据表/**/*直播观看详情*.csv'):
            df = pd.read_csv(f, skiprows=1)
            self.考勤表出现次数 += Counter([x.strip() for x in df['用户ID']])
        for f in self.root.glob('数据表/**/*直播用户列表*.csv'):
            df = pd.read_csv(f)
            self.考勤表出现次数 += Counter([x.strip() for x in df['用户ID']])

        self.异常data = {}

        if hasattr(self, '异常处理'):
            self.异常处理()

    def get_file(self, p):
        # 返回最后一个匹配的文件
        return list(self.root.glob(p))[-1]

    def 查找用户(self, 昵称='', 手机号='', *, debug=False):
        """
        :param 昵称: 昵称可以输入一个列表，会在目标昵称、真实姓名同时做检索
        :param 手机号: 手机也可以输入一个列表，因为有些人可能报名时手机号填错，可以增加一些匹配规则
        :return:
        """

        # 1 统一输入参数格式
        def tolist(x):
            if x and not isinstance(x, list):
                x = [x]
            x = [str(a) for a in x if a]
            return x

        def check_telphone(v, refs):
            if not isinstance(v, str):
                try:
                    v = int(v)
                except:
                    pass
                v = str(v)
            for a in refs:
                if a in v:
                    return True

        昵称 = tolist(昵称)
        手机号 = tolist(手机号)
        手机号 = [f'{x}' for x in 手机号]

        # 2 查找所有可能匹配的项目
        ls = []
        for idx, x in self.用户列表.iterrows():
            logo = False
            if '真实姓名' in x:
                x['姓名'] = x['真实姓名']

            if 昵称:
                if x['昵称'] in 昵称 or x['姓名'] in 昵称:
                    logo = True
            if 手机号:
                if check_telphone(x.get('账户绑定手机号'), 手机号) or check_telphone(x.get('最近采集的手机号'), 手机号):
                    logo = True
            if logo:
                ls.append(x)

        if debug:
            print('\n'.join(map(self.用户信息摘要, ls)))
            return

        return ls

    def 用户信息摘要(self, x):
        """ 输入series类型的一个条目，输出其摘要信息 """
        ls = [f'{k}={v}' for k, v in x.items() if (not isinstance(v, float) or not math.isnan(v))]
        if x['用户ID'] in self.考勤表出现次数:
            ls.append('考勤' + str(self.考勤表出现次数[x['用户ID']]) + '次')
        return ', '.join(ls)

    def 匹配用户ID(self):

        def try2int(x):
            try:
                return int(x)
            except:
                if isinstance(x, float) and math.isnan(x):
                    return ''
                return x

        def todict(row):
            """ 将表格第i行的数据提取为字典格式
            """
            msg = {}
            for k, x in enumerate(['真实姓名', '微信昵称', '手机号', '错误手机号']):
                msg[x] = ws.cell2(row, x).value
            return msg

        ws = self.wb['报名表']
        for row in tqdm(list(ws.iterrows('真实姓名')), desc='匹配进度'):
            x = todict(row)
            待查手机号 = [x['手机号']]
            t = try2int(x['错误手机号'])
            if t:
                待查手机号.append(t)
            ls = self.查找用户([x['真实姓名'], x['微信昵称']], 待查手机号)
            摘要ls = list(map(self.用户信息摘要, ls))

            用户ID = ''
            if len(ls) == 1:
                # 只有一个关联的直接匹配上，否则填空
                用户ID = ls[0]['用户ID']
            else:
                # 只有一条有考勤记录时
                flags = [('考勤' in text) for text in 摘要ls]
                if sum(flags) == 1:
                    idx = flags.index(1)
                    用户ID = ls[idx]['用户ID']
            if 用户ID == '':
                flags = [('账号状态=正常' in text) for text in 摘要ls]
                if sum(flags) == 1:
                    idx = flags.index(1)
                    用户ID = ls[idx]['用户ID']
            ws.cell2(row, '用户ID', 用户ID)
            ws.cell2(row, '参考信息', '\n'.join(摘要ls))
            row += 1
        self.wb.save(self.表格路径)

    def 获取每日统计表(self):
        ls = []
        columns = ['课次', '用户ID', '观看日期', '在线时长(分钟)']
        for f in self.root.glob('数据表/*.csv'):
            if m := re.match(r'(\d{4}\-\d{2}\-\d{2}).+?课.*?(\d+).+?直播观看详情', f.stem):
                stat_day, 课次 = date.fromisoformat(m.group(1)), int(m.group(2))
                skiprows = 1
            elif m := re.search(r'\d+届.+?(\d+).+?直播用户列表.+?(\d{4}\-\d{2}\-\d{2})', f.stem):
                stat_day, 课次 = date.fromisoformat(m.group(2)), int(m.group(1))
                skiprows = 0
            else:
                continue

            if stat_day > self.today:  # 超过self.today的数据不记录
                continue
            观看日期 = (stat_day - self.开课日期).days - 课次 + 1
            df = pd.read_csv(f, skiprows=skiprows)
            for idx, r in df.iterrows():
                k = '累计观看时长(秒)' if 观看日期 else '直播观看时长(秒)'
                ls.append([课次, r['用户ID'].strip(), 观看日期, int(int(r[k]) / 60)])

        # ext: 异常数据修正，trick
        for k, v in self.异常data.items():
            课次, 用户ID = k
            # 观看日期和在线时长可以乱填，主要是把这个用户课次记录下，后面遍历才会处理到
            ls.append([int(re.search(r'\d+', 课次).group()), 用户ID, 0, 0])

        df = pd.DataFrame.from_records(ls, columns=columns)
        return df

    def 获取考勤表中出现的所有用户名(self):
        df = self.获取每日统计表()
        print(*set(df['用户ID']), sep='\n')

    def 修订(self, 学号, 课次, 完成标记, *args):
        try:
            user_id = self.ws.cell2({'学号': 学号}, '用户ID').value
        except ValueError:
            return
        if isinstance(课次, int):
            课次 = f'第{课次:02}课'
        self.异常data[课次, user_id] = 完成标记

    def 更新统计表(self, df):
        # 1 辅助函数
        ws = self.ws
        id2row = {}
        cel = ws.findcel('用户ID').down()
        while cel.value:
            id2row[cel.value] = cel.row
            cel = cel.down()

        col = ws.findcol('第01课') - 1

        def 完成标记(items):
            """ 输入某个用户某个课次每天的累计观看时长 """
            x = {t['观看日期']: t['在线时长(分钟)'] for idx, t in items.iterrows()}
            for k in sorted(x.keys()):
                if x[k] >= self.达标时长:
                    return f'第{k}天回放' if k else '完成当堂学习'
            return f'不足{self.达标时长}分钟'

        def write(课次, 用户ID, value):
            # 优先使用修订的标记
            value = self.异常data.get((f'第{课次:02}课', 用户ID), value)
            cel = ws.cell2(id2row[用户ID], f'第{课次:02}课')
            if cel is None:
                return
            cel.value = value
            color = None
            if '完成当堂学习' in value:
                color = RgbFormatter.from_name('鲜绿色')
            elif '回放' in value:
                color = RgbFormatter.from_name('黄色')
                v1 = self.视频返款[0]
                v2 = self.视频返款[int(re.search(r'第(\d+)天', value).group(1))]
                if v2:
                    color = color.light((v1 - v2) / v2)  # 根据返款额度自动变浅
                else:  # 如果无返款额度
                    color = RgbFormatter.from_name('灰色')
            elif 课次 <= self.当天课次 - len(self.视频返款) + 1:
                cel.value = '未完成学习'
                color = RgbFormatter.from_name('红色')

            if color:
                cel.fill_color(color)

        # 2 遍历更新考勤情况数据
        for [课次, 用户ID], items in df.groupby(['课次', '用户ID']):
            if 用户ID not in id2row:
                continue
            write(课次, 用户ID, 完成标记(items))

        # 3 处理剩余用户
        for i in range(1, min(self.课次数 + 1, self.结束课次 + 1)):
            for k, v in id2row.items():
                t = ws.cell(v, col + i).value
                if not t or t in ('未开始学习', f'不足{self.达标时长}分钟'):
                    write(i, k, '未完成学习')
        for i in range(max(self.结束课次 + 1, 1), min(self.课次数 + 1, self.当天课次 + 1)):
            for k, v in id2row.items():
                if not ws.cell(v, col + i).value:
                    write(i, k, '未开始学习')

    def 生成今日返款文件(self, msg):
        """ 生成今日返款文件，及缺勤报告

        生成的csv文件共4列：订单号，反馈额，备注说明，重复校验码
        第4列可以设置一个64长度以内的标识来判重，防止重复退款
            因为每个学员每个课次理论上应该只有一次退款操作
            所以我使用"{订单号}_class{课次号}"来生成校验码

        注意，涉及给人看的数据，大部分课次编号都不会加前导0，
        涉及程序、文件名对齐的，则一般都会设前导0
        """
        ls = []  # 返款汇总
        name = []  # 生成的csv文件名

        ws = self.ws
        _col = ws.findcol('第01课') - 1

        # 1 初期要详细提示
        if self.当天课次 < 4:
            msg.append('第1次建议大家都查看下完整考勤表，记一下自己的学号，方便以后核对考勤数据。'
                       '以后群里统一以学号的方式汇报异常的考勤数据，未列出则都是"正常完成当堂学习"的学员。')
            msg.append('如果发现自己考勤数据不对，可以群里、私聊反馈给我修正。')

        # 2 要过期的课程
        if 0 < self.结束课次 + 1 <= self.课次数:
            msg.append(f'第{self.结束课次 + 1}课回放、打卡第{len(self.视频返款) - 1}天，还未完成学习的同学们请抓紧时间。')

        # 3 回放
        回放课次 = self.当天课次 - self.回放返款天数
        if 0 < 回放课次 <= self.课次数:
            title = f'第{回放课次}课回放'
            msg.append('【' + title + '名单】')
            if 回放课次 == 1:
                msg[-1] += f'（第1课仍有{len(self.视频返款) - self.回放返款天数 - 1}天可以回放）'

            d = {}
            for i in range(1, self.回放返款天数 + 1):
                d[f'第{i}天回放'] = []
            d['未完成学习'] = []

            col = _col + 回放课次
            for r in ws.iterrows('用户ID'):
                v = ws.cell(r, col).value
                if v and '回放' in v:
                    t = int(re.search(r'第(\d+)天', v).group(1))
                    if self.视频返款[t]:
                        订单号 = ws.cell2(r, '交易订单号').value
                        if 订单号:
                            cols = [订单号, self.视频返款[t],
                                    f'{self.返款标题}第{回放课次}课第{t}天完成回放',
                                    f'{订单号}_class{回放课次:02}']  # 防止重复返款的校验码
                            ls.append(','.join(map(str, cols)))
                    else:
                        v = '未完成学习'
                elif v in (f'不足{self.达标时长}分钟', '未开始学习'):
                    v = '未完成学习'
                if v and v != '完成当堂学习':
                    d[v].append(ws.cell2(r, '学号').value)

            is_empty = True
            for k, v in d.items():
                if v:
                    m = re.search(r'\d+', k)
                    t = int(m.group()) if m else 5
                    msg[-1] += f'\n{k}(返{self.视频返款[t]}元): ' + ','.join(map(str, v))
                    is_empty = False
            if is_empty:
                msg[-1] = f'第{回放课次}课因当堂满勤，无回放名单'
            else:
                name.append(f'第{回放课次}课回放')

        # 4 当堂
        if self.当天课次 <= self.课次数:
            title = f'第{self.当天课次}课当堂'
            msg.append('【' + title + '缺勤名单】')
            if self.当天课次 == 1:
                msg[-1] += '（注意只统计早晨直播情况，今日的回放数据明天会更新）'
            # d = {f'不足{self.达标时长}分钟': [], '未开始学习': []}
            d = []

            col = _col + self.当天课次
            for r in ws.iterrows('用户ID'):
                v = ws.cell(r, col).value
                if v == '完成当堂学习':
                    订单号 = ws.cell2(r, '交易订单号').value
                    if 订单号:
                        cols = [订单号, self.视频返款[0],
                                f'{self.返款标题}第{self.当天课次}课完成当堂学习',
                                f'{订单号}_class{self.当天课次:02}']  # 防止重复返款的校验码
                        ls.append(','.join(map(str, cols)))
                else:
                    d.append(ws.cell(r, 1).value)

            if d:
                msg[-1] += '\n' + ','.join(map(str, d))
            else:
                msg[-1] = title + '满勤！！！'
            name.append(f'第{self.当天课次}课当堂')

        # 5 第22课的处理
        if self.觉观禅课 and self.当天课次 == 23:
            name.append('第22课')
            for r in ws.iterrows('用户ID'):
                v = ws.cell2(r, '交易订单号').value
                if v:
                    cols = [v, self.视频返款[0],
                            f'{self.返款标题}第{self.当天课次 - 1}课',
                            f'{v}_class22']  # 防止重复返款的校验码
                    ls.append(','.join(map(str, cols)))

        # 6 返款文件
        if ls:
            ls = [x for x in ls if ('订单号' not in x and 'None' not in x)]
            (self.root / (f'第{self.当天课次:02}天 ' + '+'.join(name) + '返款.csv')).write_text('\n'.join(ls))

        # 7 提示信息
        if name:
            msg.append('+'.join(name) + '促学金已返款，同学们请查收。')
        if self.觉观禅课 and self.当天课次 == 22:
            msg.append('今晚第22课答疑不考勤，明天统一返款。')
        return msg

    def 计算打卡返款(self, 打卡次数):
        from bisect import bisect_right

        ks = [0] + list(self.打卡返款.keys())
        ks.sort()

        def find_le(a, x):
            'Find rightmost value less than or equal to x'
            i = bisect_right(a, x)
            if i:
                return a[i - 1]
            raise ValueError

        k = find_le(ks, 打卡次数)
        return self.打卡返款.get(k, 0)

    def 打卡统计(self, msg):
        """ 不知道平台有每个课程的打卡次数，自己用另外一个途径暴力搞的，220226周六15:59 现在应该没用了"""
        from openpyxl.styles import PatternFill

        # 1 至少两个目录才统计
        dirs = list((self.root / '数据表').glob_dirs('用户学习统计*'))
        if len(dirs) < 2:
            return

        # 2 读取数据
        def merge_data(d):
            """ 输入目录，汇总目录下所有xlsx文件的打卡数据

            导出的打卡数据文件尽量小些，不然这个操作速度有点慢。。。
            """
            data = {}
            for f in d.glob('*.xlsx'):
                if f.stem.startswith('~$'):
                    continue
                df = pd.read_excel(f, header=1)
                for idx, x in df.iterrows():
                    user_id = x['用户ID\t']
                    if user_id == user_id:  # nan数值的特性，这个条件不成立
                        value = x['提交打卡\t']
                        data[user_id.strip()] = int(value) if value == value else 0
                else:
                    continue

            return data

        d1, d2 = dirs[0], dirs[-1]
        data1 = merge_data(d1)
        data2 = merge_data(d2)

        # 3 d1数据标准化
        # 最理想的是d1恰好是开课前一天的统计数据，那么d2和d1直接相减即可
        # 如果d1提前了，无法判断具体打卡日期，所以还是直接相减，只能都算上
        # 如果d1延后了，那么要对d1的数据进行预处理，要减掉对应天数，但最低为0，不能为负数
        def parse_date(name):
            """ 输入文件名，获得其日期 """
            m = re.search(r'(\d{2})(\d{2})(\d{2})', name)
            year, month, day = m.groups()
            year = '20' + year
            return date(int(year), int(month), int(day))

        delta = (parse_date(d1.stem) - self.开课日期).days + 1
        if delta > 0:
            data1 = {k: max(0, v - delta) for k, v in data1.items()}

        # 4 计算新的打卡次数
        # 打卡数是有可能超过21次的
        data3 = {k: v - data1.get(k, 0) for k, v in data2.items()}

        # 5 将打卡次数写入表格
        ls = []  # 返款汇总
        ws = self.ws
        color0 = [(255, 0, 0), (255, 255, 128), (255, 255, 0), (0, 255, 0)]
        for i in ws.iterrows('用户ID'):
            c1 = ws.cell2(i, ['打卡返款', '打卡数'])
            v = c1.value = data3.get(ws.cell2(i, '用户ID').value, 0)
            v //= 5
            if v:
                money = self.打卡返款[min(v - 1, 2)]
                订单号 = ws.cell2(i, '交易订单号').value
                cols = [订单号, money, f'{self.返款标题}返学修日志促学金', f'{订单号}_journal']
                ls.append(','.join(map(str, cols)))
            else:
                money = 0

            c2 = ws.cell2(i, ['打卡返款', '返款'])
            c2.value = money

            color = RgbFormatter(*color0[min(3, v)])
            c1.fill = PatternFill(fgColor=color.hex[-6:], fill_type="solid")
            c2.fill = PatternFill(fgColor=color.hex[-6:], fill_type="solid")

        # 并计算当前的返款额
        desc = '/'.join(map(str, self.打卡返款))
        if self.当天课次 == 25:
            msg.append('已生成截止目前的打卡数据，同学们可以预先核对下，明天最后更新打卡数据后返款。'
                       f'打卡达到"5/10/15"次，依次返回"{desc}"元。'
                       '注：因技术原因，打卡数据无法精确计算，统计遵循宁可多算但无漏算的原则，所以部分同学打卡数会超过21次。')
        elif self.当天课次 == self.课次数 + len(self.视频返款) - 1:
            # 生成打卡返款
            msg.append('已完成学修日志（打卡）促学金的返款。'
                       f'打卡达到"5/10/15"次，依次返回"{desc}"元。')
            (self.root / '学修日志返款.csv').write_text('\n'.join(ls))

    def 打卡统计2(self, msg):
        from openpyxl.styles import PatternFill

        # 1 有打卡统计表才计算
        files = list((self.root / '数据表').glob_files('*打卡*.csv'))
        if len(files) < 1:
            return

        # 2 读取数据
        # df = pd.read_csv(files[-1])
        # data = {}
        # for idx, row in df.iterrows():
        #     data[row['用户id']] = row['打卡次数']

        df = None
        try:
            df = pd.read_excel(files[-1])  # 220804周四08:28，小鹅通更新了模板
        except ValueError:
            pass

        if df is None:
            try:
                df = pd.read_csv(files[-1])  # 221005周三09:19，小鹅通又双叒更新了
            except UnicodeDecodeError:
                pass

        if df is None:
            raise ValueError

        df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)
        df.columns = df.columns.map(lambda x: x.strip() if isinstance(x, str) else x)

        # data = Counter([x for x in df['用户id']])
        try:
            data = {row['用户id']: row['打卡次数'] for _, row in df.iterrows()}
        except KeyError:  # 230202周四19:49，另一种实际是xlsx格式，然后再转出csv的情况
            data = Counter([row['user_id'] for _, row in df.iterrows()])

        # 3 将打卡次数写入表格
        ls = []  # 返款汇总
        ws = self.ws
        最高返款额 = max(self.打卡返款.values())
        for i in ws.iterrows('用户ID'):
            c1 = ws.cell2(i, ['打卡返款', '打卡数'])
            打卡次数 = c1.value = data.get(ws.cell2(i, '用户ID').value, 0)
            返款额 = self.计算打卡返款(打卡次数)

            if 返款额 > 0:
                订单号 = ws.cell2(i, '交易订单号').value
                cols = [订单号, 返款额, f'{self.返款标题}返学修日志促学金', f'{订单号}_journal']
                ls.append(','.join(map(str, cols)))

            c2 = ws.cell2(i, ['打卡返款', '返款'])
            c2.value = 返款额

            if 返款额 <= 0:
                color = RgbFormatter(255, 0, 0)
            elif 返款额 < 最高返款额:
                color = RgbFormatter(255, 255, 0).light(1 - 返款额 / 最高返款额)
            elif 返款额 == 最高返款额:  # 最高返款额
                color = RgbFormatter(0, 255, 0)

            c1.fill = PatternFill(fgColor=color.hex[-6:], fill_type="solid")
            c2.fill = PatternFill(fgColor=color.hex[-6:], fill_type="solid")

        if ls:
            ls = [x for x in ls if '无订单号' not in x]

        # 4 生成通知，及返款文件
        desc1 = '/'.join(map(str, self.打卡返款.keys()))
        desc2 = '/'.join(map(str, self.打卡返款.values()))
        if self.结束课次 == self.课次数 - 1:
            msg.append('已生成截止目前的打卡数据，同学们可以预先核对下，明天最后更新打卡数据后返款。'
                       f'打卡达到"{desc1}"次，依次返回"{desc2}"元。')
        elif self.结束课次 == self.课次数:
            # 生成打卡返款
            msg.append('已完成学修日志（打卡）促学金的返款。'
                       f'打卡达到"{desc1}"次，依次返回"{desc2}"元。')
            (self.root / '学修日志返款.csv').write_text('\n'.join(ls))

    def write_返款总计(self):
        ws = self.ws
        name2idx = {k: i for i, k in enumerate(['完成当堂学习', '第1天回放', '第2天回放', '第3天回放', '第4天回放'])}
        for i in ws.iterrows('用户ID'):
            total = 0
            for k in range(1, 22):
                cel = ws.cell2(i, f'第{k:02}课')
                if not cel:
                    continue
                v = cel.value
                if v in name2idx:
                    idx = name2idx[v]
                    if idx and k > self.当天课次 - self.回放返款天数:  # 还没返款的回放金额
                        continue
                    total += self.视频返款[idx]
            v2 = ws.cell2(i, ['打卡返款', '返款']).value
            if isinstance(v2, int):
                total += v2
            if self.觉观禅课 and self.当天课次 > 22:
                total += self.视频返款[0]
            ws.cell2(i, ['已返款', '总计'], total)

    def 考勤日报(self, debug=False, journal=False):
        msg = []  # 报告内容

        # 1 日常返款
        df = self.获取每日统计表()
        self.更新统计表(df)
        self.生成今日返款文件(msg)

        # 2 打卡返款
        if journal or self.结束课次 == self.课次数:
            self.打卡统计2(msg)
        self.write_返款总计()

        # 3 结课
        if self.当天课次 == (self.课次数 + len(self.视频返款) - 1):
            msg.append('考勤返款工作已全部完成，若对自己促学金还有疑问的近日可以再跟我反馈。')

        # 4 输出日报
        if len(msg) > 1:
            msg = '\n'.join([f'{i}、{x}' for i, x in enumerate(msg, start=1)])  # 编号
        else:
            msg = msg[0]
        if self.在线表格:
            msg = f'【完整考勤表】{self.在线表格}\n' + msg
        print(msg)

        # 5 展示或保存表格内容
        if debug:
            browser.html(self.ws.to_html())
        else:
            # self.wb.save(self.表格路径)
            self.wb.save(self.root / (XlPath(self.表格路径).stem + '.xlsx'))

    def cmd(self):
        """ 初始化类后不关闭，继续轮询执行功能 """
        cmd = input('>')
        while cmd != 'exit':
            fire.Fire(self, cmd)
            cmd = input('>')

    def __call__(self, *args, **kwargs):
        print('程序测试通过，可正常使用')

    def 表格内容对齐(self, ws_name1, ws_name2, col_name):
        wb = self.wb
        wb.merge_sheets_by_keycol([wb[ws_name1], wb[ws_name2]], col_name)

        # 保存。一般不用重新再加载self.wb、self.ws，因为这里需求本来基本都是要分段重新执行程序的。
        wb.save(self.表格路径)

    def __2_自动浏览网页(self):
        pass

    def ensure_driver(self):
        if not self.driver:
            self.driver = XlChrome()
        return self.driver

    def 登录小鹅通(self, name, passwd):
        # 登录小鹅通
        driver = self.ensure_driver()
        driver.get('https://admin.xiaoe-tech.com/t/login#/acount')
        driver.locate('//*[@id="common_template_mounted_el_container"]'
                      '/div/div[1]/div[3]/div/div[4]/div/div[1]/div[1]/div/div[2]/input').send_keys(name)
        driver.locate('//*[@id="common_template_mounted_el_container"]'
                      '/div/div[1]/div[3]/div/div[4]/div/div[1]/div[2]/div/div/input').send_keys(passwd)
        driver.click('//*[@id="common_template_mounted_el_container"]/div/div[1]/div[3]/div/div[4]/div/div[2]')

        # 然后自己手动操作验证码
        # 以及选择"店铺"

    def 登录微信支付(self):
        driver = self.ensure_driver()
        driver.get('https://pay.weixin.qq.com/index.php/core/home/login')

    def 下载课次考勤数据(self, 起始课=None, 终止课=None, 文件名前缀=''):
        if 起始课 is None:
            起始课 = max(1, self.结束课次)
        if 终止课 is None:
            终止课 = min(self.当天课次, self.课次数)

        # 1 遍历课程下载表格
        driver = self.driver
        for i in range(起始课, 终止课 + 1):
            print(f'第{i}课', self.课程链接[i])
            driver.get('https://admin.xiaoe-tech.com/t/data_center/index')  # 必须要找个过渡页，不然不会更新课程链接
            driver.get(self.课程链接[i])
            driver.locate_text('//*[@id="app"]/div/div/div[1]/div[2]/div[1]/div[2]', f'第{i}课')  # 检查页面

            driver.click('//*[@id="tab-studentTab"]/span')  # 直播间用户
            driver.click('//*[@id="pane-studentTab"]/div/div[2]/div[2]/form/div[2]/button[2]/span/span')  # 导出列表
            driver.click('//*[@id="data-export-container"]/div/div[2]/div/div[2]/div[2]/button[2]/span/span')  # 导出
            time.sleep(1)

        # 2 下载表格
        # 2.1 等待下载文件生成完毕
        while True:
            driver.get('https://admin.xiaoe-tech.com/t/basic-platform/downloadCenter')
            if '任务撤回' in driver.locate('//*[@id="downloadsCenter"]/div[4]/div/div[4]/div[2]/table/tbody/tr[1]/'
                                       'td[9]/div/div/span').text:
                time.sleep(1)
            else:
                time.sleep(2)  # 不能下载太快，还是要稍微等一会
                break

        # 2.2 下载表格
        files = []
        for i in range(1, 终止课 - 起始课 + 2):
            driver.click(f'//*[@id="downloadsCenter"]/div[4]/div/div[4]/div[2]/table/tbody/tr[{i}]/'
                         f'td[9]/div/div/span[1]')
            file = driver.locate(f'//*[@id="downloadsCenter"]/div[4]/div/div[3]/table/tbody/tr[{i}]/'
                                 f'td[1]/div/div/div').text + '.csv'  # 下载后还会再多一个csv后缀
            files.append(file)
            time.sleep(1)

        # 2.3 移动表格
        time.sleep(1)
        # 这样应该也不是很泛用，有的人浏览器下载可能会放到各种自定义的其他地方
        download_path = XlPath(os.path.join(os.environ['USERPROFILE'], 'Downloads'))
        for file in files:
            src_file = (download_path / file)
            while True:
                if src_file.is_file():
                    src_file.move(self.root / '数据表' / (文件名前缀 + file), if_exists='replace')
                    break
                else:
                    time.sleep(1)

    def 批量退款(self):
        self.driver.get('https://pay.weixin.qq.com/index.php/xphp/cbatchrefund/batch_refund#/pages/index/index')

    def 申请单条退款(self, 凭证号, 退款金额=0, 退款原因=''):
        driver = self.ensure_driver()
        driver.get('https://pay.weixin.qq.com/index.php/core/refundapply')
        driver.locate('//*[@id="app"]/div/div[2]/div[2]/div[2]/div/span/input').send_keys(凭证号)
        driver.click('//*[@id="applyRefundBtn"]')  # 申请退款
        driver.locate('//*[@id="app"]/div/div[2]/div[2]/div[3]/div[2]/div/div[1]/div/span[1]/input').send_keys(
            str(退款金额))
        driver.locate('//*[@id="textInput"]').send_keys(退款原因)
        # driver.click('//*[@id="commitRefundApplyBtn"]')  # 建议手动点"提交申请"


class KqDb(Connection):
    """ 五一身心行修考勤工具 """

    def __init__(self, dbfile='5034.db', wbpath='考勤.xlsx', *args, **kwargs):
        super().__init__(dbfile, *args, **kwargs)
        p = XlPath(wbpath)
        self.wb = openpyxl.load_workbook(p)
        self.outwb = p.with_name(p.stem + '+' + p.suffix)
        # 开营第几天
        self.days = (datetime.datetime.today() - datetime.datetime(2022, 4, 29)).days

    def update_小鹅通数据(self):
        # 1 观看记录
        def add_one_csv(path, 课次名=None):
            """ 添加一个csv文件的数据 """
            # 1 确定表格存在
            if not self.has_table('观看记录'):
                cols = ['课次名 text', '用户ID text', '用户昵称 text',
                        '直播间停留秒数 integet', '累计观看秒数 integer', '直播观看秒数 integer',
                        '回放观看秒数 integer',
                        '首次进入时间 text', '最近进入时间 text', '记录时间点 text']
                cols = ', '.join(cols)
                self.execute(f'CREATE TABLE 观看记录 ({cols}, PRIMARY KEY (课次名, 用户ID, 直播间停留秒数))')

            # 2 解析csv的内容到sql中
            if not 课次名:
                课次名 = re.search(r'《(.+?)》', str(path)).group(1)
                if 课次名 == '测试链接2':
                    课次名 = '2022五一线上觉观营-0429开营'

            time_tag = datetime.datetime.fromtimestamp(os.stat(path).st_ctime).strftime('%Y-%m-%d %H:%M:%S')
            df = pd.read_csv(path, skiprows=1)
            for idx, row in df.iterrows():
                d = {'课次名': 课次名,
                     '用户ID': row['用户ID'].strip(),
                     '用户昵称': row['用户昵称'],
                     '直播间停留秒数': row['直播间停留时长(秒)'],
                     '累计观看秒数': row['累计观看时长(秒)'],
                     '直播观看秒数': row['直播观看时长(秒)'],
                     '回放观看秒数': row['回放观看时长(秒)'],
                     '首次进入时间': row['首次进入时间'].strip(),
                     '最近进入时间': row['最近进入时间'].strip(),
                     '记录时间点': time_tag,
                     }
                self.insert_row('观看记录', d)
            self.commit()

        for f in XlPath('数据表').glob('*直播观看详情*.csv'):
            add_one_csv(f)

        # 2 打卡记录
        self.execute('DROP TABLE IF EXISTS 打卡记录')
        # 时间点最新的数据
        f = list(XlPath('数据表').glob('*打卡日记*.csv'))[-1]
        df = pd.read_excel(f)
        df.columns = df.columns.str.replace('\t', '')  # 删除列名里的\t
        df = df.replace('\t', '', regex=True)  # 删除值里的\t
        df.to_sql('打卡记录', con=self)

    def update_用户列表(self, path):
        """ 更新用户列表数据 """
        # 1 删除旧表，创建新表
        self.execute('DROP TABLE IF EXISTS 用户列表')
        # 需要预取一下数据，知道所有字段
        df = pd.read_csv(path)
        cols = {k: 'text' for k in df.columns}
        cols['年龄'] = 'integer'
        desc = ','.join([f'{k} {v}' for k, v in cols.items()])
        self.execute(f'CREATE TABLE 用户列表 ({desc}, PRIMARY KEY (用户ID))')
        self.commit()

        # 2 插入每一行数据
        for idx, row in df.iterrows():
            for k in ['账户绑定手机号', '最近采集手机号']:
                # 有的是写成公式的'=手机号'
                if isinstance(row[k], str) and '=' in row[k]:
                    row[k] = re.search(r'\d+', row[k]).group()
            self.insert_row('用户列表', row.to_dict())
        self.commit()

    def 用户信息摘要(self, x):
        """ 输入series类型的一个条目，输出其摘要信息 """
        ls = [f'{k}={v}' for k, v in x.items() if (not isinstance(v, float) or not math.isnan(v))]
        v = x["用户ID"].strip()
        # if v == 'u_615026b5b5db9_0uLTQndbsC':
        #     print(v)
        n = self.execute(f'SELECT COUNT(*) FROM 观看记录 WHERE 用户ID="{v}"').fetchone()[0]
        if n:
            ls.append(f'考勤{n}次')
        return ', '.join(ls)

    def 查找用户(self, 昵称='', 手机号='', *, debug=False):
        """
        :param 昵称: 昵称可以输入一个列表，会在目标昵称、真实姓名同时做检索
        :param 手机号: 手机也可以输入一个列表，因为有些人可能报名时手机号填错，可以增加一些匹配规则
        :return:
        """

        # 1 统一输入参数格式
        def tolist(x):
            if x and not isinstance(x, list):
                x = [x]
            x = [str(a) for a in x if a]
            return x

        def check_telphone(v, refs):
            if not isinstance(v, str):
                try:
                    v = int(v)
                except:
                    pass
                v = str(v)
            for a in refs:
                if a in v:
                    return True

        昵称 = tolist(昵称)
        手机号 = tolist(手机号)
        手机号 = [f'{x}' for x in 手机号]

        # 2 查找所有可能匹配的项目
        ls = []
        for x in self.exec_dict('SELECT * FROM 用户列表'):
            logo = False
            if 昵称:
                if x['昵称'] in 昵称 or x['真实姓名'] in 昵称:
                    logo = True
            if 手机号:
                if check_telphone(x['账户绑定手机号'], 手机号) or check_telphone(x['最近采集手机号'], 手机号):
                    logo = True
            if logo:
                ls.append(x)

        if debug:
            print('\n'.join(map(self.用户信息摘要, ls)))
            return

        return ls

    def 匹配用户ID(self, wbpath, sheet='报名表'):
        p = XlPath(wbpath)
        wb = openpyxl.load_workbook(wbpath)
        ws = wb[sheet]

        def try2int(x):
            try:
                return int(x)
            except:
                if isinstance(x, float) and math.isnan(x):
                    return ''
                return x

        def todict(row):
            """ 将表格第i行的数据提取为字典格式
            """
            msg = {}
            for k, x in enumerate(['姓名', '微信昵称', '手机', '微信']):
                msg[x] = ws.cell2(row, x).value
            return msg

        for row in tqdm(list(ws.iterrows('姓名')), desc='匹配进度'):
            x = todict(row)
            待查手机号 = [x['手机']]
            t = try2int(x['微信'])
            if t:
                待查手机号.append(t)
            ls = self.查找用户([x['姓名'], x['微信昵称']], 待查手机号)
            # 只有一个关联的直接匹配上，否则填空
            ws.cell2(row, '用户ID', ls[0]['用户ID'] if len(ls) == 1 else '')
            ws.cell2(row, '参考信息', '\n'.join(map(self.用户信息摘要, ls)))
            row += 1

            # if len(ls) > 1 and ls[1]['用户ID'] == 'u_615026b5b5db9_0uLTQndbsC':
            #     break

        wb.save(str(p.with_name(p.stem + '+' + p.suffix)))

    def get_user_观看打卡记录(self, user_id):
        """ 获得一个用户的所有观看记录情况 """

        # 1 观看记录
        classes = [
            ['2022五一线上觉观营-0429开营', '2022-04-29 19:00', '2022-04-29 20:58', 0],
            ['4.30第一堂', '2022-04-30 05:20', '2022-04-30 06:27', 420],
            ['4.30第二堂', '2022-04-30 08:00', '2022-04-30 10:40', 2400],
            # 张国莉师兄说以时间表为准，所以我适当放宽。以15:30结束为准，结束前3分钟退出不算早退。
            ['4.30第三堂', '2022-04-30 13:30', '2022-04-30 15:33', 180],
            ['4.30第四堂', '2022-04-30 19:30', '2022-04-30 21:03', 180],
            ['5.1第一堂', '2022-05-01 05:20', '2022-05-01 06:17', 0],
            ['5.1第二堂', '2022-05-01 08:00', '2022-05-01 10:13', 13 * 60],
            ['5.1第三堂', '2022-05-01 13:30', '2022-05-01 15:11', 0],
            ['5.1第四堂', '2022-05-01 19:30', '2022-05-01 21:00', 0],
            ['5.2第一堂', '2022-05-02 05:20', '2022-05-02 06:03', 0],
            ['5.2第二堂', '2022-05-02 08:00', '2022-05-02 10:07', 7 * 60],
            ['5.2第三堂', '2022-05-02 13:30', '2022-05-02 15:28', 0],
            ['5.2第四堂', '2022-05-02 19:30', '2022-05-02 21:02', 2 * 60],
            ['5.3第一堂', '2022-05-03 05:20', '2022-05-03 06:11', 0],
            ['5.3第二堂', '2022-05-03 08:00', '2022-05-03 10:12', 12 * 60],
            ['5.3第三堂', '2022-05-03 13:30', '2022-05-03 15:43', 13 * 60],
            ['5.3第四堂', '2022-05-03 19:30', '2022-05-03 20:57', 0],
            ['5.4第一堂', '2022-05-04 05:20', '2022-05-04 06:22', 2 * 60],
            ['5.4第二堂', '2022-05-04 08:00', '2022-05-04 09:39', 0],
            ['5.4第三堂', '2022-05-04 13:30', '2022-05-04 15:43', 13 * 60],
            ['5.4第四堂', '2022-05-04 19:30', '2022-05-04 21:02', 2 * 60],
            ['5.5第一堂', '2022-05-05 05:20', '2022-05-05 06:14', 0],
            ['5.5第二堂', '2022-05-05 08:00', '2022-05-05 10:05', 5 * 60],
            ['5.5第三堂', '2022-05-05 13:30', '2022-05-05 15:21', 0],
            ['5.5第四堂', '2022-05-05 19:30', '2022-05-05 21:02', 2 * 60],
            ['5.6第一堂', '2022-05-06 05:20', '2022-05-06 06:23', 3 * 60],
            ['5.6第二堂', '2022-05-06 08:00', '2022-05-06 10:15', 15 * 60],
            ['5.6第三堂', '2022-05-06 13:30', '2022-05-06 15:26', 0],
        ]

        def 课次message(name):
            for k, item in enumerate(classes):
                if item[0] == name:
                    da = datetime.datetime.fromisoformat(item[1])
                    db = datetime.datetime.fromisoformat(item[2])
                    return k, da, db, item[3]

        items = self.exec_dict(f'SELECT * FROM 观看记录 WHERE 用户ID="{user_id}"')
        ls = ['缺勤'] * len(classes) + [''] * (36 - len(classes))
        for x in items:
            k, da, db, bias = 课次message(x['课次名'])

            # d1: 提早进入直播间的，顺推至标准开课时间
            d1 = datetime.datetime.fromisoformat(x['首次进入时间'])
            if d1 < da:
                if d1 <= datetime.datetime(2022, 5, 4) and user_id != 'u_612d861be8fde_BfDyBtACTv':
                    # 5月4日之前的，缺勤做了大放水~~ 提前登录的，可以换好多正课时长~~ 但是李娟的单独调整
                    x['直播观看秒数'] += (da - d1).seconds
                d1 = da

            # d2: 进入时间，加上直播间观察时间作为退出时间
            d2 = d1 + datetime.timedelta(seconds=x['直播观看秒数'])
            # 结束时间不超过db
            if d2 > db:
                d2 = db

            desc = ''
            if max((d1 - da).seconds, 0) + max((db - d2).seconds, 0) >= (1800 + bias):
                desc = ',缺勤'
            else:
                if (d1 - da).seconds > 59:
                    desc = ',迟到'

            # 直播结束才进入的
            if d1 >= db:
                ls[k] = '缺勤'
            else:
                ls[k] = d1.strftime('%H:%M') + '/' + d2.strftime('%H:%M') + desc
        观看记录 = ls

        if user_id == 'u_61ad24c13ec15_qBIqwSvejE':  # 1组占萍
            观看记录[0] = '19:09/20:58,迟到'

        # 2 打卡记录
        def brief_time(x, ref):
            """ x相对ref的简化时间显示 """
            t = int((x - ref).seconds / 60)
            tag = f'{t // 60:02}:{t % 60:02}'
            if t < 0:
                tag = '-' + tag
            return tag

        打卡记录 = [''] * 9
        start_day = datetime.datetime(2022, 4, 29)
        for k, tag in enumerate(['立下学修目标', '第一天', '第二天', '第三天', '第四天', '第五天', '第六天', '第七天']):
            x = self.execute(f'SELECT 打卡时间 FROM 打卡记录 WHERE user_id="{user_id}" AND 所属主题="{tag}"').fetchall()
            x = [a[0] for a in x]  # 只有一列数据
            if not x:
                打卡记录[k] = '未打卡'
            else:
                x = sorted(x)  # 排序后分别处理
                d1 = start_day + datetime.timedelta(days=k)
                d2 = start_day + datetime.timedelta(days=k, hours=19)
                打卡记录[k] = ','.join([brief_time(datetime.datetime.fromisoformat(a), d1) for a in x])
                if datetime.datetime.fromisoformat(x[0]) > d2:
                    打卡记录[k] += ',迟打'

        # 3 观看和打卡数据合并
        用户记录 = [观看记录[0], 打卡记录[0]]
        for i in range(1, 9):
            b = i * 4 + 1
            用户记录 += 观看记录[b - 4:b] + [打卡记录[i]]

        return 用户记录

    def update_wb(self):
        """ 更新工作薄考勤内容 """
        ws = self.wb['考勤表']
        # 1 考勤数据
        for i in ws.iterrows('用户ID'):
            user_id = ws.cell2(i, '用户ID').value
            msg = self.get_user_观看打卡记录(user_id)
            for j, v in enumerate(msg, start=10):
                if v:
                    c = ws.cell(i, j)
                    v0 = c.value
                    if v0:  # 如果有特殊标记，需要修正处理
                        m = re.search(grp_chinese_char(), v)  # 判断原有v中是否有中文
                        if m:  # 如果v有中文，需要替换为v0的标记
                            v = re.sub(grp_chinese_char() + '+', v0, v)
                        elif v:  # 如果v没有中文，直接在加上特殊标记后缀
                            v += f',{v0}'
                        else:  # 如果v也没有标记，保留v0原始标记内容
                            v = v0
                    c.value = v
                    if '上班' in v or '工作' in v:
                        color = '鲜红'
                    elif '请假' in v:
                        color = '灰色'
                    elif '迟到' in v or '迟打' in v:
                        color = '鲜黄'
                    elif '缺勤' in v or '未打卡' in v:
                        color = '鲜红'
                    else:
                        color = '鲜绿色'
                    c.fill_color(color)

        # 2 剩余促学金
        for i in ws.iterrows('用户ID'):
            money = 1000
            累计缺勤天数 = 0
            for t in range(1, min(8, (datetime.datetime.today() - datetime.datetime(2022, 4, 28)).days)):  # 第几天
                logo = False
                # 每天上课情况
                迟到_num, 缺勤_num, 请假_num = 0, 0, 0
                for jj in range(1, 5):
                    j = t * 5 + 6 + jj
                    v = ws.cell(i, j).value
                    if not v:
                        v = ''
                    if '迟到' in v:
                        迟到_num += 1
                    elif '缺勤' in v or '上班' in v or '工作' in v:
                        缺勤_num += 1
                    elif '请假' in v:
                        请假_num += 1
                if 迟到_num + 缺勤_num >= 3:
                    money -= 240
                    logo = True
                else:
                    money -= 迟到_num * 25
                    money -= 缺勤_num * 60
                    # money -= 请假_num * 20

                if 缺勤_num:
                    累计缺勤天数 += 1
                    if 累计缺勤天数 >= 3:
                        money -= 1000
                else:
                    累计缺勤天数 = 0

                # 每天打卡情况
                v = ws.cell(i, t * 5 + 11).value
                if v and not logo:
                    if '迟打' in v:
                        money -= 30
                    elif '未打卡' in v:
                        money -= 60
            ws.cell2(i, ['促学金', '剩余']).value = max(0, money)

        # 3 给整行加上分割线
        from openpyxl.styles import Border, Side, Alignment
        for i in ws.iterrows('用户ID'):
            v = ws.cell(i, 1).value
            if isinstance(v, str) and '组' in v:
                side = Side(border_style='medium', color='000000')
                border = Border(top=side)
                for j in range(1, 50):
                    ws.cell(i, j).border = border
                    if j > 9:
                        ws.cell(i, j).alignment = Alignment(horizontal='left', vertical='center')

        # 4 各组综合情况
        # 4.1 各组打卡情况
        打卡天数 = 6
        ls = []
        title = ''  # 组名
        columns = ['分组', '姓名', '日期', '打卡状态']
        for i in ws.iterrows('用户ID'):
            v = ws.cell(i, 1).value
            if isinstance(v, str) and '组' in v:
                title = f'第{int(chinese2digits(v[1:-1])):02}组'
            for j in range(打卡天数):
                打卡状态 = ws.cell(i, 16 + j * 5).value
                ls.append([title, ws.cell2(i, '姓名').value, j, 打卡状态])
        df = pd.DataFrame.from_records(ls, columns=columns)

        ls = []
        columns = ['分组', '总人数', '每天平均补打人数及比例', '每天平均未打人数及未打比例',
                   '4月30日未打', '5月1日未打', '5月2日未打', '5月3日未打', '5月4日未打', '5月5日未打']
        for title, items in df.groupby('分组'):
            n = len(items)
            a = sum(items['打卡状态'].str.contains('迟打'))
            b = sum(items['打卡状态'].str.contains('未打卡'))
            record = [title, n // 打卡天数, f'{a / 打卡天数:.1f}，{a / n:.0%}', f'{b / 打卡天数:.1f}，{b / n:.0%}']
            for j in range(打卡天数):
                msg = []
                补打, 未打 = 0, 0
                # 迟打
                # xs = items[(items['日期'] == j) & (items['打卡状态'].str.contains('迟打'))]['姓名']
                # if len(xs):
                #     补打 += len(xs)
                #     msg.append(f'补打{len(xs)}人')
                # 未打
                xs = items[(items['日期'] == j) & (items['打卡状态'].str.contains('未打卡'))]['姓名']
                if len(xs):
                    未打 += len(xs)
                    msg.append(f'{len(xs)}人：{"，".join(xs)}')
                record.append('，'.join(msg))
            ls.append(record)

        df = pd.DataFrame.from_records(ls, columns=columns)
        # browser(df)

        # 4.2 每天打卡情况
        self.wb.save(self.outwb)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import os
import subprocess

python_path = sys.executable

ori_path = os.getcwd()
os.chdir('ppocr/postprocess/pse_postprocess/pse')
if subprocess.call(
        '{} setup.py build_ext --inplace'.format(python_path), shell=True) != 0:
    raise RuntimeError(
        'Cannot compile pse: {}, if your system is windows, you need to install all the default components of `desktop development using C++` in visual studio 2019+'.
        format(os.path.dirname(os.path.realpath(__file__))))
os.chdir(ori_path)

from .pse import pse



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 22:56

"""
文本处理、常用正则匹配模式

下面大量的函数前缀含义：
grp，generate regular pattern，生成正则模式字符串
grr，generate regular replace，生成正则替换目标格式
"""

import base64
import bisect
import collections
import io
import logging
import os
import re
import sys

from pyxllib.prog.newbie import RunOnlyOnce
from pyxllib.prog.pupil import dprint
from pyxllib.text.newbie import circlednumber2digits, digits2circlednumber, roman2digits, digits2roman


def shorten(s, width=200, placeholder='...'):
    """
    :param width: 这个长度是上限，即使用placeholder时的字符串总长度也在这个范围内

    >>> shorten('aaa', 10)
    'aaa'
    >>> shorten('hell world! 0123456789 0123456789', 11)
    'hell wor...'
    >>> shorten("Hello  world!", width=12)
    'Hello world!'
    >>> shorten("Hello  world!", width=11)
    'Hello wo...'
    >>> shorten('0123456789 0123456789', 2, 'xyz')  # 自己写的shorten
    'xy'

    注意textwrap.shorten的缩略只针对空格隔开的单词有效，我这里的功能与其不太一样
    >>> textwrap.shorten('0123456789 0123456789', 11)  # 全部字符都被折叠了
    '[...]'
    >>> shorten('0123456789 0123456789', 11)  # 自己写的shorten
    '01234567...'
    """
    s = re.sub(r'\s+', ' ', str(s))
    n, m = len(s), len(placeholder)
    if n > width:
        s = s[:max(width - m, 0)] + placeholder
    return s[:width]  # 加了placeholder在特殊情况下也会超，再做个截断最保险

    # return textwrap.shorten(str(s), width)


def strfind(fullstr, objstr, *, start=None, times=0, overlap=False):
    r"""进行强大功能扩展的的字符串查找函数

    TODO 性能有待优化

    :param fullstr: 原始完整字符串
    >>> strfind('aabbaabb', 'bb')  # 函数基本用法
    2

    :param objstr: 需要查找的目标字符串，可以是一个list或tuple
    TODO 有空看下AC自动机，看这里是否可以优化提速，或者找现成的库接口
    >>> strfind('bbaaaabb', 'bb') # 查找第1次出现的位置
    0
    >>> strfind('aabbaabb', 'bb', times=1) # 查找第2次出现的位置
    6
    >>> strfind('aabbaabb', 'cc') # 不存在时返回-1
    -1
    >>> strfind('aabbaabb', ['aa', 'bb'], times=2)
    4

    :param start: 起始查找位置。默认值为0，当times<0时start的默认值为-1。
    >>> strfind('aabbaabb', 'bb', start=2) # 恰好在起始位置
    2
    >>> strfind('aabbaabb', 'bb', start=3)
    6
    >>> strfind('aabbaabb', ['aa', 'bb'], start=5)
    6

    :param times: 定位第几次出现的位置，默认值为0，即从前往后第1次出现的位置。
        如果是负数，则反向查找，并返回的是目标字符串的起始位置。
    >>> strfind('aabbaabb', 'aa', times=-1)
    4
    >>> strfind('aabbaabb', 'aa', start=5, times=-1)
    4
    >>> strfind('aabbaabb', 'aa', start=3, times=-1)
    0
    >>> strfind('aabbaabb', 'bb', start=7, times=-1)
    6

    :param overlap: 重叠情况是否重复计数
    >>> strfind('aaaa', 'aa', times=1)  # 默认不计算重叠部分
    2
    >>> strfind('aaaa', 'aa', times=1, overlap=True)
    1

    >>> strfind(r'\item=\item+', (r'\item', r'\test'), start=1)
    6
    """

    def nonnegative_min_value(*arr):
        """计算出最小非负整数，如果没有非负数，则返回-1"""
        arr = tuple(filter(lambda x: x >= 0, arr))
        return min(arr) if arr else -1

    def nonnegative_max_value(*arr):
        """计算出最大非负整数，如果没有非负数，则返回-1"""
        arr = tuple(filter(lambda x: x >= 0, arr))
        return max(arr) if arr else -1

    # 1 根据times不同，start的初始默认值设置方式也不同
    if times < 0 and start is None:
        start = len(fullstr) - 1  # 反向查找start设到末尾字符-1
    if start is None:
        start = 0  # 正向查找start设为0
    p = -1  # 记录答案位置，默认找不到

    # 2 单串匹配
    if isinstance(objstr, str):  # 单串匹配
        offset = 1 if overlap else len(objstr)  # overlap影响每次偏移量

        # A、正向查找
        if times >= 0:
            p = start - offset
            for _ in range(times + 1):
                p = fullstr.find(objstr, p + offset)
                if p == -1:
                    return -1

        # B、反向查找
        else:
            p = start + offset + 1
            for _ in range(-times):
                p = fullstr.rfind(objstr, 0, p - offset)
                if p == -1:
                    return -1

    # 3 多模式匹配（递归调用，依赖单串匹配功能）
    else:
        # A、正向查找
        if times >= 0:
            p = start - 1
            for _ in range(times + 1):
                # 把每个目标串都找一遍下一次出现的位置，取最近的一个
                #   因为只找第一次出现的位置，所以overlap参数传不传都没有影响
                # TODO 需要进行性能对比分析，有必要的话后续可以改AC自动机实现多模式匹配
                ls = tuple(map(lambda x: strfind(fullstr, x, start=p + 1, overlap=overlap), objstr))
                p = nonnegative_min_value(*ls)
                if p == -1:
                    return -1

        # B、反向查找
        else:
            p = start + 1
            for _ in range(-times):  # 需要循环处理的次数
                # 使用map对每个要查找的目标调用strfind
                ls = tuple(map(lambda x: strfind(fullstr, x, start=p - 1, times=-1, overlap=overlap), objstr))
                p = nonnegative_max_value(*ls)
                if p == -1:
                    return -1

    return p


def findspan(src, sub, start=0, end=None):
    """ str.find的封装

    :param sub:
        str，普通的字符串查找
        re.Pattern，正则模式的查找
    :return: (start, end)
        找不到的时候返回 (-1, -1)
        否则返回区间的左开右闭位置
    """
    if end is None:
        end = len(src)

    if isinstance(sub, str):
        pos = src.find(sub, start, end)
    elif isinstance(sub, re.Pattern):
        pattern = sub
        m = pattern.search(src[start:end])
        if m:
            pos = m.start() + start
            sub = m.group()
        else:
            pos = -1
    else:
        raise TypeError

    if pos == -1:
        return -1, -1
    else:
        return pos, pos + len(sub)


def substr_count(src, sub, overlape=False):
    """ 判断字符串src中符合pattern的字串有几个 """
    if overlape:
        raise NotImplementedError
    else:
        if isinstance(sub, str):
            cnt = src.count(sub)
        elif isinstance(sub, re.Pattern):
            cnt = len(sub.findall(src))
        else:
            raise TypeError

    return cnt


class Stdout:
    """重定向标准输出流，切换print标准输出位置

    使用with语法调用
    """

    def __init__(self, path=None, mode='w'):
        """
        :param path: 可选参数
            如果是一个合法的文件名，在__exit__时，会将结果写入文件
            如果不合法不报错，只是没有功能效果
        :param mode: 写入模式
            'w': 默认模式，直接覆盖写入
            'a': 追加写入
        """
        self.origin_stdout = sys.stdout
        self._path = path
        self._mode = mode
        self.strout = io.StringIO()
        self.result = None

    def __enter__(self):
        sys.stdout = self.strout
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self.origin_stdout
        self.result = str(self)

        # 如果输入的是一个合法的文件名，则将中间结果写入
        if not self._path:
            return

        try:
            with open(self._path, self._mode) as f:
                f.write(self.result)
        except TypeError as e:
            logging.exception(e)
        except FileNotFoundError as e:
            logging.exception(e)

        self.strout.close()

    def __str__(self):
        """在这个期间获得的文本内容"""
        if self.result:
            return self.result
        else:
            return self.strout.getvalue()


def int2myalphaenum(n):
    """
    :param n: 0~52的数字
    """
    if 0 <= n <= 52:
        return '_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'[n]
    else:
        print('警告：不在处理范围内的数值', n)
        raise ValueError


def ensure_gbk(s):
    """检查一个字符串的所有内容是否能正常转为gbk，
    如果不能则ignore掉不能转换的部分"""
    try:
        s.encode('gbk')
    except UnicodeEncodeError:
        origin_s = s
        s = s.encode('gbk', errors='ignore').decode('gbk')
        print('警告：字符串存在无法转为gbk的字符', origin_s, s)
    return s


def digit2weektag(d):
    """ 输入数字1~7，转为“周一~周日”

    >>> digit2weektag(1)
    '周一'
    >>> digit2weektag('7')
    '周日'
    """
    d = int(d)
    if 1 <= d <= 7:
        return '周' + '一二三四五六日'[d - 1]
    else:
        raise ValueError


def fullwidth2halfwidth(ustring):
    """ 把字符串全角转半角

    python3环境下的全角与半角转换代码和测试_大数据挖掘SparkExpert的博客-CSDN博客:
    https://blog.csdn.net/sparkexpert/article/details/82749207

    >>> fullwidth2halfwidth("你好ｐｙｔｈｏｎａｂｄａｌｄｕｉｚｘｃｖｂｎｍ")
    '你好pythonabdalduizxcvbnm'
    """
    ss = []
    for s in ustring:
        for uchar in s:
            inside_code = ord(uchar)
            if inside_code == 12288:  # 全角空格直接转换
                inside_code = 32
            elif 65281 <= inside_code <= 65374:  # 全角字符（除空格）根据关系转化
                inside_code -= 65248
            ss.append(chr(inside_code))
    return ''.join(ss)


def fullwidth2halfwidth2(ustring):
    """ 不处理标点符号的版本

    >>> fullwidth2halfwidth2("你好ｐｙｔｈｏｎａｂｄａ，ｌｄｕｉｚｘｃｖｂｎｍ")
    '你好pythonabda，lduizxcvbnm'
    """
    ss = []
    for s in ustring:
        for uchar in s:
            if uchar in '：；！（），？＂．':
                ss.append(uchar)
            else:
                inside_code = ord(uchar)
                if inside_code == 12288:  # 全角空格直接转换
                    inside_code = 32
                elif 65281 <= inside_code <= 65374:  # 全角字符（除空格）根据关系转化
                    inside_code -= 65248
                ss.append(chr(inside_code))
    return ''.join(ss)


def halfwidth2fullwidth(ustring):
    """ 把字符串全角转半角

    >>> halfwidth2fullwidth("你好ｐｙｔｈｏｎａｂｄａｌｄｕｉｚｘｃｖｂｎｍ")
    '你好ｐｙｔｈｏｎａｂｄａｌｄｕｉｚｘｃｖｂｎｍ'
    """
    ss = []
    for s in ustring:
        for uchar in s:
            inside_code = ord(uchar)
            if inside_code == 32:  # 全角空格直接转换
                inside_code = 12288
            elif 33 <= inside_code <= 126:  # 全角字符（除空格）根据关系转化
                inside_code += 65248
            ss.append(chr(inside_code))
    return ''.join(ss)


class ContentPartSpliter:
    """ 文本内容分块处理 """

    @classmethod
    def multi_blank_lines(cls, content, leastlines=2):
        """ 用多个空行隔开的情况

        :param leastlines: 最少2个空行隔开，为新的一块内容
        """
        fmt = r'\n{' + str(leastlines) + ',}'
        parts = [x.strip() for x in re.split(fmt, content)]
        parts = list(filter(bool, parts))  # 删除空行
        return parts


class ContentLine(object):
    """ 用行数的特性分析一段文本 """

    def __init__(self, content):
        """用一段文本初始化"""
        self.content = content  # 原始文本
        self.linepos = list()  # linepos[i-1] = v：第i行终止位置（\n）所在下标为v
        for i in range(len(self.content)):
            if self.content[i] == '\n':
                self.linepos.append(i)
        self.linepos.append(len(self.content))
        self.lines = self.content.splitlines()  # 每一行的文本内容

    def line_start_pos(self, line):
        """第line行的其实pos位置"""
        pass

    def lines_num(self):
        """返回总行数"""
        return self.content.count('\n')

    def match_lines(self, pattern):
        """返回符合正则规则的行号

        180515扩展： pattern也能输入一个函数
        """
        # 1 定义函数句柄
        if not callable(pattern):
            def f(s):
                return re.search(pattern, s)
        else:
            f = pattern
        # 2 循环判断
        res = list()
        for i, line in enumerate(self.lines):
            if f(line):
                res.append(i)
        return res

    def in_line(self, ob):
        """输入关键词ob，返回行号"""

        if hasattr(ob, 'span'):
            return self.in_line(ob.span()[0])
        elif isinstance(ob, int):
            "如果给入一个下标值，如23，计算第23个字符处于原文中第几行"
            return bisect.bisect_right(self.linepos, ob - 1) + 1
        elif isinstance(ob, str):
            "输入一段文本，判断该文中有哪些行与该行内容相同"
            res = list()
            for i, line in enumerate(self.lines):
                if line == ob:
                    res.append(i + 1)
            return res
        elif isinstance(ob, (list, tuple, collections.Iterable)):
            return list(map(self.in_line, ob))
        else:
            raise ValueError(f'类型错误 {type(ob)}')

    def regular_search(self, re_str):
        """同InLine，但是支持正则搜索"""
        return self.in_line(re.finditer(re_str, self.content))

    def lines_content(self, lines) -> str:
        """返回lines集合中数字所对行号的所有内容

        注意输入的lines起始编号是1
        """
        lines = sorted(set(lines))  # 去重
        res = map(lambda n: '{:6} {}'.format(n, self.lines[n - 1]), lines)
        return '\n'.join(res)

    def __str__(self):
        return self.content


def digits2chinese(n):
    """TODO：目前处理范围有限，还需要再扩展
    """
    s = '十一二三四五六七八九'
    if n == 0:
        return '零'
    elif n <= 10:
        return s[n % 10]
    elif n < 20:
        return '十' + s[n % 10]
    elif n < 100:
        return s[n // 10] + s[n % 10]
    else:
        raise NotImplementedError


def chinese2digits(chinese_str):
    """把汉字变为阿拉伯数字
    https://blog.csdn.net/leon_wzm/article/details/78963082
    """

    def inner(m):
        t = m.group()
        if t is None or t.strip() == '':
            raise ValueError(f'input error for {chinese_str}')
        t = t.strip()
        t = t.replace('百十', '百一十')
        common_used_numerals = {'零': 0, '一': 1, '二': 2, '两': 2, '三': 3, '四': 4, '五': 5,
                                '六': 6, '七': 7, '八': 8, '九': 9, '十': 10,
                                '百': 100, '千': 1000, '万': 10000, '亿': 100000000}
        total = 0
        r = 1  # right，右边一位的值
        for i in range(len(t) - 1, -1, -1):  # 从右往左一位一位读取
            val = common_used_numerals.get(t[i])  # 使用get不存在会返回None
            if val is None:
                # dprint(chinese_str)
                return chinese_str
                # raise ValueError(f't[i]={t[i]} can not be accepted.')
            if val >= 10 and i == 0:  # 最左位是“十百千万亿”这样的单位数词
                if val > r:  # 一般是“十三”这类会进入这个if分支
                    r = val
                    total += val
                else:
                    r *= val
            elif val >= 10:
                if val > r:  # 跳了单位数词（正常情况都会跳），例如 一万一百零三
                    r = val
                else:  # 单位数词叠加情况，例如 一千亿
                    r *= val
            else:  # 不是单位数词的数词，如果上一步是单位数词，增加一个单位量
                total += r * val
        return str(total)

    return re.sub(r'[零一二两三四五六七八九十百千万亿]+', inner, chinese_str)


def briefstr(s):
    """对文本内容进行一些修改，从而简化其内容，提取关键信息
    一般用于字符串近似对比
    """
    # 1 删除所有空白字符
    # debuglib.dprint(debuglib.typename(s))
    s = re.sub(r'\s+', '', s)
    # 2 转小写字符
    s = s.casefold()
    return s


@RunOnlyOnce
def grp_bracket(depth=0, left='{', right=None, inner=False):
    r"""括号匹配，默认花括号匹配，也可以改为圆括号、方括号匹配。

    效果类似于“{.*?}”，
    但是左右花括号是确保匹配的，有可选参数可以提升支持的嵌套层级，
    数字越大匹配嵌套能力越强，但是速度性能会一定程度降低。
    例如“grp_bracket(5)”。

    :param depth: 括号递归深度
    :param left: 左边字符：(、[、{
    :param right: 右边字符
    :param inner: 默认只是返回匹配的正则表达式，不编组
        如果设置inner=True，则会对括号内的内容编组
        该功能用来代替原来的BRACE5等机制

    :return:

    先了解一下正则常识：
    >>> re.sub(r'[^\[\]]', r'', r'a[b]a[]') # 删除非方括号
    '[][]'
    >>> re.sub(r'[^\(\)]', r'', r'a(b)a()') # 删除非圆括号
    '()()'
    >>> re.sub(r'[^()]', r'', r'a(b)a()') # 不用\也可以
    '()()'

    该函数使用效果：
    >>> re.sub(grp_bracket(5), r'', r'x{aaa{b{d}b}ccc{d{{}e}ff}gg}y')
    'xy'
    >>> re.sub(grp_bracket(5, '(', ')'), r'', r'x(aaa(b(d)b)ccc(d(()e)ff)gg)y')
    'xy'
    >>> re.sub(grp_bracket(5, '[', ']'), r'', r'x[aaa[b[d]b]ccc[d[[]e]ff]gg]y')
    'xy'
    """
    # 用a, b简化引用名称
    a, b = left, right
    b = b or {'(': ')', '[': ']', '{': '}'}[a]
    # 特殊符号需要转义
    if a in '([':
        a = '\\' + a
    if b in ')]':
        b = '\\' + b
    c = f'[^{a}{b}]'
    # 建立匹配素材
    pattern_0 = f'{a}{c}*{b}'
    pat_left = f'{a}(?:{c}|'
    pat_right = f')*{b}'

    # 生成匹配规则的函数
    def gen(pattern, depth=0):
        while depth:
            pattern = pat_left + pattern + pat_right
            depth -= 1
        return pattern

    s = gen(pattern_0, depth=depth)

    # inner
    if inner:
        return f'{a}({s[len(a):len(s) - len(b)]}){b}'
    else:
        return s


def grp_chinese_char():
    return r'[\u4e00-\u9fa5，。；？（）【】、①-⑨]'


def grr_check(m):
    """用来检查匹配情况"""
    s0 = m.group()
    pass  # 还没想好什么样的功能是和写到re.sub里面的repl
    return s0


def printoneline(s):
    """将输出控制在单行，适应终端大小"""
    try:
        columns = os.get_terminal_size().columns - 3  # 获取终端的窗口宽度
    except OSError:  # 如果没和终端相连，会抛出异常
        # 这应该就是在PyCharm，直接来个大值吧
        columns = 500
    s = shorten(s, columns)
    print(s)


def count_word(s, *patterns):
    """ 统计一串文本中，各种规律串出现的次数

    :param s: 文本内容
    :param patterns: （正则规则）
        匹配的多个目标模式list
        按优先级一个一个往后处理，被处理掉的部分会用\x00代替
    :return: Counter.most_common() 对象
    """
    s = str(s)

    if not patterns:  # 不写参数的时候，默认统计所有单个字符
        return collections.Counter(list(s)).most_common()

    ls = []
    for t in patterns:
        ls += re.findall(t, s)
        s = re.sub(t, '\x00', s)
        # s = re.sub(r'\x00+', '\x00', s)  # 将连续的特殊删除设为1，减短字符串长度，还未试验这段代码精确度与效率
    ct = collections.Counter(ls)

    ls = ct.most_common()
    for i in range(len(ls)):
        ls[i] = (ls[i][1], repr(ls[i][0])[1:-1])
    return ls


class Base85Coder:
    """base85编码、解码器

    对明文，加密/编码/encode 后已经是乱了看不懂，但是对这个结果还要二次转义
    对乱码，解密/解码/decode 时顺序要反正来，先处理二次转义，再处理base85

    使用示例：
    key = 'xV~>Y|@muL<UK$*agCQp=t4c0R_y`Z2;q%s?o8S9(3D5W^-NA&}6v){Twj7MzGePJEfik1bBhn!d#I+HlXFOr'
    coder = Base85Coder(key)
    b = coder.encode('陈坤泽 abc')
    dprint(b)  # b<str>=d@7;B}ww?}zfGP#;1
    s = coder.decode(b)
    dprint(s)  # s<str>=陈坤泽 abc
    """
    DEFAULT_KEY = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!#$%&()*+-;<=>?@^_`{|}~'
    CHARS_SET = set(DEFAULT_KEY)

    def __init__(self, key=None):
        """key，允许设置密钥，必须是"""
        # 1 分析key是否合法
        if key:
            if len(key) != 85 or set(key) != Base85Coder.CHARS_SET:
                dprint(key)  # 输入key无效
                key = None
        self.key = key

        # 2 制作转换表 trantab
        if key:
            self.encode_trantab = str.maketrans(Base85Coder.DEFAULT_KEY, key)
            self.decode_trantab = str.maketrans(key, Base85Coder.DEFAULT_KEY)
        else:
            self.encode_trantab = self.decode_trantab = None

    def encode(self, s):
        """将字符串转字节"""
        b = base64.b85encode(s.encode('utf8'))
        b = str(b)[2:-1]
        if self.encode_trantab:
            b = b.translate(self.encode_trantab)
        return b

    def decode(self, b):
        if self.decode_trantab:
            b = b.translate(self.decode_trantab)
        b = b.encode('ascii')
        s = base64.b85decode(b).decode('utf8')
        return s


def check_text_row_column(s):
    """对一段文本s，用换行符分割行，用至少4个空格或\t分割列，分析数据的行、列数
    :return:
        (n, m)，每列的列数相等，则会返回n、m>=0的tuple
        (m1, m2, ...)，如果有列数不相等，则会返回每行的列数组成的tuple
            每个元素用负值代表不匹配
    """
    # 拆开每行的列
    if not s: return (0, 0)
    lines = [re.sub(r'( {4,}|\t)+', r'\t', line.strip()).split('\t') for line in s.splitlines()]
    cols = [len(line) for line in lines]  # 计算每行的列数
    if min(cols) == max(cols):
        return len(lines), cols[0]
    else:
        return [-col for col in cols]


class ListingFormat:
    r"""列表格式化工具

    >>> li = ListingFormat('（1）')
    >>> li
    （1）
    >>> li.next()
    >>> li
    （2）

    >>> li = ListingFormat(('一、选择题', '二、填空题', '三、解答题'))
    >>> li
    一、选择题
    >>> li.next()
    >>> li
    二、填空题
    """
    formats = {'[零一二三四五六七八九十]+': (chinese2digits, digits2chinese),
               r'\d+': (int, str),
               '[A-Z]': (lambda x: ord(x) - ord('A') + 1, lambda x: chr(ord('A') + x - 1)),
               '[a-z]': (lambda x: ord(x) - ord('a') + 1, lambda x: chr(ord('a') + x - 1)),
               '[①②③④⑤⑥⑦⑧⑨⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳]': (circlednumber2digits, digits2circlednumber),
               '[ⅠⅡⅢⅣⅤⅥⅦⅧⅨⅩⅪⅫ]': (roman2digits, digits2roman)}

    def __init__(self, s='1'):
        """
        :param s: 列表的格式，含数值和装饰
            数值形式，目前有六种形式：一 1 A a ① Ⅰ
                起始值可以不是1，例如写'三'、'D'等
                装饰的格式，常见的有：'({})'  '（{}）'  '{}、'  '{}.' '{}. '
            list或tuple，按顺序取用，用完后不再设置前缀
        >> ListingFormat('一', '{}、')

        TODO 目前只考虑值较小的情况，如果值太大，有些情况会出bug、报错
        """
        if isinstance(s, str):
            for k, funcs in ListingFormat.formats.items():
                if re.search(k, s):
                    self.form = re.sub(k, '{}', s)
                    self.value = int(funcs[0](re.search(k, s).group()))
                    self.func = funcs[1]
                    break
            else:
                raise ValueError('列表初始化格式不对 s=' + str(s))
        elif isinstance(s, (list, tuple)):
            self.form = s
            self.value = 0
            self.func = None
        else:
            raise ValueError('列表初始化格式不对 s=' + str(s))

    def reset(self, start=1):
        """重置初始值"""
        self.value = start

    def next(self):
        self.value += 1

    def __repr__(self):
        if self.func:
            return self.form.format(self.func(self.value))
        else:
            return self.form[self.value]


class StrDiffType:
    typename = {
        0: '完全相同',
        1: '忽略case后，相同',
        2: '忽略blank后，相同',
        3: '忽略case+blank后，相同',
        4: 'dt是gt局部信息（精度ok，召回不行）',
        5: '忽略case后，dt是gt局部信息',
        6: '忽略blank后，dt是gt局部信息',
        7: '忽略case+blank后，dt是gt局部信息',
        8: 'gt是dt局部信息（召回ok，精度不行）',
        9: '忽略case后，gt是dt局部信息',
        10: '忽略blank后，gt是dt局部信息',
        11: '忽略case+blank后，gt是dt局部信息',
        16: '其他情况 （可以根据实验情况，后续继续细分类别）'
    }

    @classmethod
    def main_difftype(cls, dt, gt):
        if not dt or not gt:
            return 16
        elif dt == gt:
            return 0
        elif dt in gt:
            return 4
        elif gt in dt:
            return 8
        else:
            return 16

    @classmethod
    def difftype(cls, dt, gt):
        """ 判断两段字符串dt,gt的差异所属类别
        """
        if not dt or not gt: return 16

        t = cls.main_difftype(dt, gt)
        if t < 16:
            return t

        t = cls.main_difftype(dt.lower(), gt.lower()) + 1
        if t < 16:
            return t

        dt2, gt2 = re.sub(r'\s+', '', dt), re.sub(r'\s+', '', gt)
        t = cls.main_difftype(dt2, gt2) + 2
        if t < 16:
            return t

        dt3, gt3 = dt2.lower(), gt2.lower()
        t = cls.main_difftype(dt3, gt3) + 3
        if t < 16:
            return t
        else:
            return 16


class BookContents:
    """ 书本目录类 """

    def __init__(self):
        self.contents = []  # 目录条目按顺序保存在list中

    def add(self, level, title, page=None):
        """
        Args:
            level:
            title:
            page: 不一定要放整数的页数，也可以放其他一些比例之类的数值

        Returns:

        """
        self.contents.append([level, title, page])

    def format_numbers(self, number='normal', *, indent='', start_level=1, jump=False):
        """ 每级目录的编号

        :param number: 编号格式，目前有默认方式，以后有需要可以扩展其他模式
        :param start_level: 开始展示的层级（高层级也会展示，只是不带编号和缩进）
            可以设为负数，表示自动推算，比如-1
        :param jump: 支持跳级，比如2级"3"，跳到4级本来是"3.0.1"，但开启该参数则会优化为"3.1"
        :return: list，跟contents等长，表示每个标题的编号，可能为空''
        """
        # 1
        if start_level == -1:
            # 自动推算合适的开始编号
            # -1模式，表示第一个不只一项的level
            levels = [x[0] for x in self.contents]
            levels_cnt = collections.Counter(levels)
            for i in range(min(levels), max(levels) + 1):
                if levels_cnt[i] > 1:
                    start_level = i
                    break

        # 2
        ls = []
        ct = collections.defaultdict(int)
        for x in self.contents:
            # print(x)
            level = x[0]
            sign = indent * (level - start_level)

            # 处理计数器
            ct[level] += 1
            for k, v in ct.items():
                if k > level:
                    ct[k] = 0

            # 当前编号
            if number == 'normal':
                numbers = [ct[i] for i in range(start_level, level + 1)]
                if jump:  # 过滤0
                    numbers = [x for x in numbers if x]
                sign += '.'.join(map(str, numbers))
            else:
                pass

            ls.append(sign)

        return ls

    def format_str(self, indent='\t', *, number='normal', page=False, start_level=1, jump=False):
        """ 转文本展示

        :param indent: 每级展示的缩进量
        :param page: 是否展示页码
        """
        numbers = self.format_numbers(number, indent=indent, start_level=start_level, jump=jump)

        # 2
        ls = []
        for num, x in zip(numbers, self.contents):
            level, title, page_ = x

            # 标题
            if level < start_level:
                sign = title
            else:
                sign = ' '.join([num, x[1]])

            # 加后缀
            if page:
                sign += f'，{page_}'

            ls.append(sign)

        return '\n'.join(ls)


def continuous_zero(s):
    """ 返回一个字符串中连续0的位置

    :param s: 一个字符串

    做html转latex表格中，合并单元格的处理要用到这个函数计算cline

    >>> continuous_zero('0100')  # 从0开始编号，左闭右开区间
    [(0, 1), (2, 4)]
    """
    return [m.span() for m in re.finditer(r'0+', s)]
